#ifndef BRUTE_FORCE_H
#define BRUTE_FORCE_H

#include "data_structs.h"


void calc_accs(particle** pcont, int n, params* par);

void calc_accs_particles(particle* p1, particle* p2, params* par);


#endif